/*
 * This software is in the public domain under CC0 1.0 Universal plus a 
 * Grant of Patent License.
 * 
 * To the extent possible under law, the author(s) have dedicated all
 * copyright and related and neighboring rights to this software to the
 * public domain worldwide. This software is distributed without any
 * warranty.
 * 
 * You should have received a copy of the CC0 Public Domain Dedication
 * along with this software (see the LICENSE.md file). If not, see
 * <http://creativecommons.org/publicdomain/zero/1.0/>.
 */
package org.moqui.impl.screen

import groovy.transform.CompileStatic
import org.codehaus.groovy.runtime.InvokerHelper
import org.moqui.context.ExecutionContext
import org.moqui.impl.actions.XmlAction
import org.moqui.impl.context.ContextBinding
import org.moqui.impl.context.ExecutionContextFactoryImpl
import org.moqui.context.ContextStack

import org.slf4j.Logger
import org.slf4j.LoggerFactory

class ScreenSection {
    protected final static Logger logger = LoggerFactory.getLogger(ScreenSection.class)

    protected Node sectionNode
    protected String location

    protected Class conditionClass = null
    protected XmlAction condition = null
    protected XmlAction actions = null
    protected ScreenWidgets widgets = null
    protected ScreenWidgets failWidgets = null

    ScreenSection(ExecutionContextFactoryImpl ecfi, Node sectionNode, String location) {
        this.sectionNode = sectionNode
        this.location = location

        // prep condition attribute
        String conditionAttr = sectionNode."@condition"
        if (conditionAttr) conditionClass = new GroovyClassLoader().parseClass(conditionAttr)

        // prep condition element
        if (sectionNode.condition && sectionNode.condition[0].children()) {
            // the script is effectively the first child of the condition element
            condition = new XmlAction(ecfi, (Node) sectionNode."condition"[0].children()[0], location + ".condition")
        }
        // prep actions
        if (sectionNode.actions) {
            actions = new XmlAction(ecfi, (Node) sectionNode."actions"[0], location + ".actions")
            // if (location.contains("FOO")) logger.warn("====== Actions for ${location}: ${actions.writeGroovyWithLines()}")
        }
        // prep widgets
        if (sectionNode.widgets) widgets = new ScreenWidgets((Node) sectionNode."widgets"[0], location + ".widgets")
        // prep fail-widgets
        if (sectionNode."fail-widgets") {
            failWidgets = new ScreenWidgets((Node) sectionNode."fail-widgets"[0], location + ".fail-widgets")
        }
    }

    @CompileStatic
    void render(ScreenRenderImpl sri) {
        ContextStack cs = (ContextStack) sri.ec.context
        if (sectionNode.name() == "section-iterate") {
            // if nothing to iterate over, all done
            def list = sri.ec.resource.expression(sectionNode["@list"] as String, null)
            if (!list) {
                if (logger.traceEnabled) logger.trace("Target list [${list}] is empty, not rendering section-iterate at [${location}]")
                return
            }
            Iterator listIterator = null
            if (list instanceof Iterator) listIterator = (Iterator) list
            else if (list instanceof Map) listIterator = ((Map) list).entrySet().iterator()
            else if (list instanceof Iterable) listIterator = ((Iterable) list).iterator()

            // TODO: handle paginate, paginate-size (lower priority...)
            int index = 0
            while (listIterator != null && listIterator.hasNext()) {
                Object entry = listIterator.next()
                try {
                    cs.push()

                    cs.put((String) sectionNode["@entry"], (entry instanceof Map.Entry ? entry.getValue() : entry))
                    if (sectionNode["@key"] && entry instanceof Map.Entry)
                        cs.put((String) sectionNode["@key"], entry.getKey())

                    cs.put("sectionEntryIndex", index)
                    cs.put(((String) sectionNode["@entry"]) + "_index", index)
                    cs.put(((String) sectionNode["@entry"]) + "_has_next", listIterator.hasNext())

                    renderSingle(sri)
                } finally {
                    cs.pop()
                }
                index++
            }
        } else {
            // NOTE: don't push/pop context for normal sections, for root section want to be able to share-scope when it
            // is included by another screen so that fields set will be in context of other screen
            renderSingle(sri)
        }
    }

    @CompileStatic
    protected void renderSingle(ScreenRenderImpl sri) {
        if (logger.traceEnabled) logger.trace("Begin rendering screen section at [${location}]")
        ExecutionContext ec = sri.getEc()
        boolean conditionPassed = true
        if (condition != null) conditionPassed = condition.checkCondition(ec)
        if (conditionPassed && conditionClass != null) {
            Script script = InvokerHelper.createScript(conditionClass, new ContextBinding(ec.getContext()))
            Object result = script.run()
            conditionPassed = result as boolean
        }

        if (conditionPassed) {
            if (actions != null) actions.run(ec)
            if (widgets != null) widgets.render(sri)
        } else {
            if (failWidgets != null) failWidgets.render(sri)
        }
        if (logger.traceEnabled) logger.trace("End rendering screen section at [${location}]")
    }
}
